#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<netinet/in.h>
#include<arpa/nameser.h>
#include<resolv.h>
#include<errno.h>
#include<signal.h>
#include<netdb.h>
#include<unistd.h>
#include<ctype.h>
#include<pthread.h>
#include"dnsqry.h"

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
extern pthread_mutex_t initex;

#define NS_VERSION_STRING "VERSION.BIND."

void *vcheck(void *ptr)
{
    struct funcargs *da = (struct funcargs *) ptr;
    register unsigned char *p;
    unsigned char *op;
    struct servers *llp = da->reserved;
    struct hostent *host, myhost;
    int count, sendlen;
    char *ns;
    union {
	HEADER hdr;
	char buf[PACKETSZ];
    } response;
    /* Use a copy of the variable containing the nameserver address so the
     * mutual exclusion lock may be unlocked for the duration of the
     * gethostbyname() call, which could block for a considerable amount of 
     * time.
     */
    if (!(ns = (char *) malloc((sendlen = strlen(llp->ns)) + 1)))
	vexit("malloc");
    strncpy(ns, llp->ns, sendlen);
    *(ns + sendlen) = 0;
#ifdef DEBUG
    tdpf("pthread_mutex_unlock(&initex)");
#endif
    pthread_mutex_unlock(&initex);
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    /* Exclude gethostbyname() calls from each other since the function is
       is not reetrant. */
#ifdef DEBUG
    tdpf("pthread_mutex_lock(&mutex)");
#endif
    pthread_mutex_lock(&mutex);
#ifdef DEBUG
    tdpf("gethostbyname");
#endif
    if (!(host = gethostbyname(ns))) {
#ifdef DEBUG
	tdpf("pthread_mutex_unlock(&mutex)");
#endif
	pthread_mutex_unlock(&mutex);
	printf("Unable to resolve: %s\n", ns);
	free(ns);
	puts("Possible erreneous NS record.");
#ifdef DEBUG
	tdpf("pthread_exit");
#endif
	pthread_exit((void *) -1);
    }
#ifdef DEBUG
    tdpf("pthread_mutex_unlock(&mutex)");
#endif
    pthread_mutex_unlock(&mutex);
    free(ns);
    memcpy((void *) &myhost, (void *) host,
	   (size_t) sizeof(struct hostent));
    memcpy((void *) &_res.nsaddr_list[0].sin_addr,
	   (void *) host->h_addr_list[0], (size_t) host->h_length);
    spin();
#ifdef DEBUG
    tdpf("res_query");
#endif
    if (
	(sendlen =
	 res_query(NS_VERSION_STRING, C_CHAOS, T_TXT,
		   (unsigned char *) &response, sizeof(response))) < 0) {
#ifdef DEBUG
	tdpf("pthread_exit");
#endif
	pthread_exit((void *) -1);
    }
#ifdef DEBUG
    tdpf("pthread_mutex_unlock");
#endif
    if (!axfrtest(&myhost, da->dom))
	printf("%s appears willing to zone transfer %s\n",
	       llp->ns, da->dom);
    spin();
    da->eom = response.buf + sendlen;
    da->cp = response.buf + sizeof(HEADER);
    da->cp += skipname(response.buf, da->cp, da->eom) + QFIXEDSZ;
    da->bufp = response.buf;
    da->cp += skiptodata(da);
    if (da->cp < da->eom) {
	if (da->type == T_TXT) {
	    op = p = da->cp + 1;
	    if (!(llp->ver = (char *) calloc(MAXDNAME + 1, sizeof(char))))
		vexit("malloc");
	    do {
		if (!*p || !isprint(*p) || isspace(*p) || !isdigit(*op)) {
		    printf("%s returned unusual version string", llp->ns);
		    /* Initially, all version strings were going to be printed
		     * by default, but I noticed some _very_ odd version
		     * responses while testing; one server returned vt100  
		     * escape characters that resembled those of flash.c! */
		    if (da->force)
			printf(" (%s)", op);
		    *llp->ver = -1;
		    printf("\n");
#ifdef DEBUG
		    tdpf("pthread_exit");
#endif
		    pthread_exit((void *) -1);
		}
		p++;
	    } while (p != da->eom);
	    count = p - op;
	    /* Paranoid strncpy(). */
	    strncpy(llp->ver, op, count);
	    /* BTW, there's no need to null-terminate the previous string;
	     * it was calloc()'ed. */
	    if ((p = strchr(llp->ver, '\n')))
		*p = 0;
	    printf("%s is %s\n", llp->ns, llp->ver);
	    /* Exit function -- relevant information has been
	     * recorded.
	     */
#ifdef DEBUG
	    tdpf("pthread_exit");
#endif
	    pthread_exit((void *) 0);
	} else {
	    puts("Possible Malformed DNS response");
	}
    }
#ifdef DEBUG
    tdpf("pthread_exit");
#endif
    pthread_exit((void *) 0);
}
